#include <petscsys.h>
#include <petscvec.h>
#include <petscksp.h>
#include <petsctao.h>
#include <petscmat.h>


//--------------------------------------------------------------------------------
int main(int argc, char **args)
{

    PetscErrorCode ierr;
    MPI_Comm       comm;
    Mat            H, Schur, A01, A10;
    Vec            rhs, sol, tmp1, tmp2;
    KSP            ksp, ksp_schur;
    PC             pc;
    PetscInt       size;
    const PetscInt nH = 10, nrhs = 1;

    ierr = PetscInitialize(&argc, &args, NULL, NULL);

    comm = PETSC_COMM_WORLD;
    ierr = MPI_Comm_size(comm, &size); CHKERRMPI(ierr);
    if(size > 1)
        SETERRQ(comm, PETSC_ERR_WRONG_MPI_SIZE, "Incorrect number of processors");

    /* Vecs */
    ierr = VecCreateSeq(comm, nrhs, &rhs); CHKERRQ(ierr);
    ierr = VecSet(rhs, 1.);                CHKERRQ(ierr);
    ierr = VecDuplicate(rhs, &sol);        CHKERRQ(ierr);
    ierr = VecCreateSeq(comm, nH, &tmp1);  CHKERRQ(ierr);
    ierr = VecDuplicate(tmp1, &tmp2);      CHKERRQ(ierr);

    /* KSP; need to do that here so MATLMVM is registered */
    ierr = KSPCreate(comm, &ksp);        CHKERRQ(ierr);
    ierr = KSPSetType(ksp, KSPGMRES);    CHKERRQ(ierr);
    ierr = KSPGetPC(ksp, &pc);           CHKERRQ(ierr);
    ierr = PCSetType(pc, PCNONE);        CHKERRQ(ierr);
    ierr = KSPGMRESSetRestart(ksp, 500); CHKERRQ(ierr);
    ierr = KSPGMRESSetOrthogonalization(ksp, KSPGMRESModifiedGramSchmidtOrthogonalization); CHKERRQ(ierr);

    /* Hessian approximation */
    ierr = MatCreate(comm, &H);                      CHKERRQ(ierr);
    ierr = MatSetSizes(H, nH, nH, nH, nH);               CHKERRQ(ierr);
    ierr = MatSetType(H, MATLMVMBFGS);               CHKERRQ(ierr);
    ierr = MatLMVMAllocate(H, tmp1, tmp2); CHKERRQ(ierr);
    // ierr = MatSetUp(H);                              CHKERRQ(ierr);

    /* Schur complement */
    ierr = MatCreateSeqAIJ(comm, 1, nH, 1, NULL, &A10); CHKERRQ(ierr);
    ierr = MatSetUp(A10);                               CHKERRQ(ierr);
    ierr = MatSetValue(A10, 0, 0, 1., INSERT_VALUES);   CHKERRQ(ierr);
    ierr = MatAssemblyBegin(A10, MAT_FINAL_ASSEMBLY);   CHKERRQ(ierr);
    ierr = MatAssemblyEnd(A10, MAT_FINAL_ASSEMBLY);     CHKERRQ(ierr);

    ierr = MatCreateTranspose(A10, &A01); CHKERRQ(ierr);

    ierr = MatCreateSchurComplement(H,
                                   H,
                                   A01,
                                   A10,
                                   NULL,
                                   &Schur); CHKERRQ(ierr);

    ierr = MatSchurComplementGetKSP(Schur, &ksp_schur); CHKERRQ(ierr);
    ierr = KSPSetType(ksp_schur, KSPPREONLY);           CHKERRQ(ierr);
    ierr = KSPGetPC(ksp_schur, &pc);                    CHKERRQ(ierr);
    ierr = PCSetType(pc, PCLMVM);                       CHKERRQ(ierr);
    ierr = PCLMVMSetMatLMVM(pc, H);                     CHKERRQ(ierr);

    ierr = KSPSetOperators(ksp, Schur, Schur); CHKERRQ(ierr);
    ierr = KSPSetUp(ksp);                      CHKERRQ(ierr);

    /* Solve */
    ierr = KSPSolve(ksp, rhs, sol); CHKERRQ(ierr);

    ierr = VecDestroy(&rhs);  CHKERRQ(ierr);
    ierr = VecDestroy(&sol);  CHKERRQ(ierr);
    ierr = VecDestroy(&tmp1); CHKERRQ(ierr);
    ierr = VecDestroy(&tmp2); CHKERRQ(ierr);

    ierr = MatDestroy(&H);     CHKERRQ(ierr);
    ierr = MatDestroy(&Schur); CHKERRQ(ierr);
    ierr = MatDestroy(&A01);   CHKERRQ(ierr);
    ierr = MatDestroy(&A10);   CHKERRQ(ierr);
    ierr = KSPDestroy(&ksp);   CHKERRQ(ierr);
    PetscFinalize();

    return 0;
}
